/*
 * Copyright (c) 2008, Your Corporation. All Rights Reserved.
 */

package au.net.ocean.maven.plugin.extractor.util;

import org.apache.maven.artifact.DependencyResolutionRequiredException;
import org.apache.maven.project.MavenProject;

import java.io.DataInputStream;
import java.io.File;
import java.io.IOException;
import java.security.AccessControlContext;
import java.security.AccessController;
import java.security.PrivilegedActionException;
import java.security.PrivilegedExceptionAction;

/**
 * Uses the magic of ClassPath and ClassFile to hide the details of reading class bytecode from JARs or directory files.
 *
 * @author <a href="mailto:dave.whitla@ocean.net.au">Dave Whitla</a>
 * @version $Id: ProjectCompileClassPathLoader.java 0 27/03/2008 15:04:40 dwhitla $
 */
public class ProjectCompileClassPathLoader extends ClassLoader {

    static final char PACKAGE_SEPARATOR_CHAR = '.';
    /* The context to be used when loading classes and resources */
    private AccessControlContext acc;
    private ClassPath classPath;

    /**
     * Constructs a new ProjectCompileClassPathLoader for the given MavenProject. The project compile classpath
     * will be searched in the order specified for classes and resources after first
     * searching in the specified parent class loader.
     * <p/>
     * <p>If there is a security manager, this method first calls the security manager's <code>checkCreateClassLoader</code>
     * method to ensure creation of a class loader is allowed.
     *
     * @param project the MavenProject who's compile classpath is to be searched for classes and resources
     * @param parent the parent class loader for delegation
     * @throws SecurityException if a security manager exists and its <code>checkCreateClassLoader</code> method doesn't allow
     *                           creation of a class loader.
     * @see SecurityManager#checkCreateClassLoader
     */
    public ProjectCompileClassPathLoader(MavenProject project, ClassLoader parent) throws DependencyResolutionRequiredException {
        super(parent);
        // this is to make the stack depth consistent with 1.1
        SecurityManager securityManager = System.getSecurityManager();
        if (securityManager != null) {
            securityManager.checkCreateClassLoader();
        }
        setProject(project);
        acc = AccessController.getContext();
    }

    private void setProject(MavenProject project) throws DependencyResolutionRequiredException {
        StringBuilder classPathBuilder = new StringBuilder();
        // build the path to avoid bug #6610897 in the ClassPath(String[]) constructor
        for (Object classPathElement : project.getCompileClasspathElements()) {
            if (classPathBuilder.length() > 0) {
                classPathBuilder.append(File.pathSeparatorChar);
            }
            classPathBuilder.append(classPathElement);
        }
        classPath = new ClassPath(classPathBuilder.toString());
    }


    /**
     * Constructs a new ProjectCompileClassPathLoader for the given MavenProject using the
     * default delegation parent <code>ClassLoader</code>. The project compile classpath
     * will be searched in the order returned by <code>Project.getCompileClasspathElements()</code>
     * for classes and resources after first searching in the specified parent class loader.
     * <p/>
     * <p>If there is a security manager, this method first calls the security manager's
     * <code>checkCreateClassLoader</code> method to ensure creation of a class loader is allowed.
     *
     * @param project the MavenProject who's compile classpath is to be searched for classes and resources
     * @throws SecurityException if a security manager exists and its <code>checkCreateClassLoader</code> method doesn't allow
     *                           creation of a class loader.
     * @see SecurityManager#checkCreateClassLoader
     */
    public ProjectCompileClassPathLoader(MavenProject project) throws DependencyResolutionRequiredException {
        super();
        // this is to make the stack depth consistent with 1.1
        SecurityManager securityManager = System.getSecurityManager();
        if (securityManager != null) {
            securityManager.checkCreateClassLoader();
        }
        setProject(project);
        acc = AccessController.getContext();
    }


    /**
     * Finds the class with the specified <a href="#name">binary name</a> from the MavenProject
     * compile classpath. Any classpath URLs referring to JAR files are loaded and opened as needed
     * until the class is found.
     *
     * @param name The <a href="#name">binary name</a> of the class
     * @return The resulting <tt>Class</tt> object
     * @throws ClassNotFoundException If the class could not be found
     */
    @Override
    protected Class<?> findClass(final String name) throws ClassNotFoundException {
        try {
            return AccessController.doPrivileged(new PrivilegedExceptionAction<Class>() {
                public Class<?> run() throws ClassNotFoundException {
                    byte[] byteCode = loadClassData(name);
                    return defineClass(name, byteCode, 0, byteCode.length, getClass().getProtectionDomain());
                }
            }, acc);
        } catch (PrivilegedActionException pae) {
            throw (ClassNotFoundException) pae.getException();
        }
    }

    /**
     * Load the class with the given fully qualified name from the ClassPath.
     * Borrowed from sun.tools.java.ClassPathLoader because it is not available in all JDKs
     */
    private byte[] loadClassData(String className) throws ClassNotFoundException {
        // Build the file name and subdirectory from the class name
        String filename = className.replace('.', File.separatorChar) + ".class";

        // Have ClassPath find the file for us, and wrap it in a ClassFile.
        // Note:  This is where it looks inside jar files that are specified in the path.
        ClassFile classFile = classPath.getFile(filename);

        if (classFile != null) {
            // Provide the most specific reason for failure in addition to ClassNotFound in the event of an error
            Exception rootCause = null;
            byte data[] = null;

            try {
                // ClassFile shields us from knowing if the class is a separate file or an entry in a jar file.
                DataInputStream input = new DataInputStream(classFile.getInputStream());

                // Can't rely on input available() since it will be something unusual if it's a jar file!  May need
                // to worry about a possible problem if someone makes a jar file entry with a size greater than max int.
                data = new byte[(int) classFile.length()];
                try {
                    input.readFully(data);
                } catch (IOException e) {
                    // Something actually went wrong reading the file.  This is a real error so save it to report it.
                    data = null;
                    rootCause = e;
                } finally {
                    try {
                        input.close();
                    } catch (IOException e) {
                        // Don't care if there's an exception on close! Lame that close can throw an IOException anyway
                    }
                }
            } catch (IOException e) {
                // Couldn't get the input stream for the file.  This is probably also a real error.
                rootCause = e;
            }

            if (data == null) {
                throw new ClassNotFoundException(className, rootCause);
            }
            return data;
        }

        // Couldn't find the file in the class path.
        throw new ClassNotFoundException(className);
    }

    /**
     * Creates a new instance of ProjectCompileClassPathLoader for the specified
     * MavenProject and parent class loader. If a security manager is
     * installed, the <code>loadClass</code> method of the ProjectCompileClassPathLoader
     * returned by this method will invoke the
     * <code>SecurityManager.checkPackageAccess</code> method before
     * loading the class.
     *
     * @param project the MavenProject who's compile classpath is to be searched for classes and resources
     * @param parent the parent class loader for delegation
     * @return the resulting class loader
     */
    public static ProjectCompileClassPathLoader newInstance(final MavenProject project, final ClassLoader parent)
            throws DependencyResolutionRequiredException {
        // Save the caller's context
        AccessControlContext acc = AccessController.getContext();
        // Need a privileged block to create the class loader
        ProjectCompileClassPathLoader loader;
        try {
            loader = AccessController.doPrivileged(new PrivilegedExceptionAction<ProjectCompileClassPathLoader>() {
                public ProjectCompileClassPathLoader run() throws DependencyResolutionRequiredException {
                    return new FactoryProjectCompileClassPathLoader(project, parent);
                }
            });
        } catch (PrivilegedActionException e) {
            throw (DependencyResolutionRequiredException) e.getException();
        }
        // Now set the context on the loader using the one we saved,
        // not the one inside the privileged block...
        loader.acc = acc;
        return loader;
    }

    /**
     * Creates a new instance of ProjectCompileClassPathLoader for the specified
     * MavenProject and default parent class loader. If a security manager is
     * installed, the <code>loadClass</code> method of the ProjectCompileClassPathLoader
     * returned by this method will invoke the
     * <code>SecurityManager.checkPackageAccess</code> method before
     * loading the class.
     *
     * @param project the MavenProject who's compile classpath is to be searched for classes and resources
     * @return the resulting class loader
     */
    public static ProjectCompileClassPathLoader newInstance(final MavenProject project) throws DependencyResolutionRequiredException {
        // Save the caller's context
        AccessControlContext acc = AccessController.getContext();
        // Need a privileged block to create the class loader
        ProjectCompileClassPathLoader loader;
        try {
            loader = AccessController.doPrivileged(new PrivilegedExceptionAction<ProjectCompileClassPathLoader>() {
                public ProjectCompileClassPathLoader run() throws DependencyResolutionRequiredException {
                    return new FactoryProjectCompileClassPathLoader(project);
                }
            });
        } catch (PrivilegedActionException e) {
            throw (DependencyResolutionRequiredException) e.getException();
        }

        // Now set the context on the loader using the one we saved,
        // not the one inside the privileged block...
        loader.acc = acc;
        return loader;
    }

    private static final class FactoryProjectCompileClassPathLoader extends ProjectCompileClassPathLoader {

        FactoryProjectCompileClassPathLoader(MavenProject project, ClassLoader parent) throws DependencyResolutionRequiredException {
            super(project, parent);
        }

        FactoryProjectCompileClassPathLoader(MavenProject project) throws DependencyResolutionRequiredException {
            super(project);
        }

        public final synchronized Class loadClass(String name, boolean resolve) throws ClassNotFoundException {
            // First check if we have permission to access the package. This
            // should go away once we've added support for exported packages.
            SecurityManager securityManager = System.getSecurityManager();
            if (securityManager != null) {
                int i = name.lastIndexOf(PACKAGE_SEPARATOR_CHAR);
                if (i != -1) {
                    securityManager.checkPackageAccess(name.substring(0, i));
                }
            }
            return super.loadClass(name, resolve);
        }
    }

}

